@@ -1,4 +1,5 @@
 """Utility helpers to simplify working with yaml-based data."""
+
 # pylint: disable=too-many-lines
 from __future__ import annotations
 
@@ -6,68 +7,56 @@ import functools
 import logging
 import os
 import re
+from collections.abc import Callable, Iterator, Sequence
 from io import StringIO
-from typing import (
-    TYPE_CHECKING,
-    Any,
-    Callable,
-    Dict,
-    Iterator,
-    Pattern,
-    Tuple,
-    Union,
-    cast,
-)
+from pathlib import Path
+from re import Pattern
+from typing import TYPE_CHECKING, Any, cast
 
 import ruamel.yaml.events
 from ruamel.yaml.comments import CommentedMap, CommentedSeq, Format
+from ruamel.yaml.composer import ComposerError
 from ruamel.yaml.constructor import RoundTripConstructor
 from ruamel.yaml.emitter import Emitter, ScalarAnalysis
 
 # Module 'ruamel.yaml' does not explicitly export attribute 'YAML'; implicit reexport disabled
 # To make the type checkers happy, we import from ruamel.yaml.main instead.
 from ruamel.yaml.main import YAML
-from ruamel.yaml.nodes import ScalarNode
-from ruamel.yaml.representer import RoundTripRepresenter
-from ruamel.yaml.scalarint import ScalarInt
-from ruamel.yaml.tokens import CommentToken
+from ruamel.yaml.parser import ParserError
+from ruamel.yaml.scalarint import HexInt, ScalarInt
 from yamllint.config import YamlLintConfig
 
 from ansiblelint.constants import (
     ANNOTATION_KEYS,
     NESTED_TASK_KEYS,
     PLAYBOOK_TASK_KEYWORDS,
-    SKIPPED_RULES_KEY,
 )
-from ansiblelint.errors import MatchError
-from ansiblelint.file_utils import Lintable
-from ansiblelint.utils import get_action_tasks, normalize_task
+from ansiblelint.utils import Task
 
 if TYPE_CHECKING:
     # noinspection PyProtectedMember
-    from ruamel.yaml.comments import LineCol  # pylint: disable=ungrouped-imports
+    from ruamel.yaml.comments import LineCol
+    from ruamel.yaml.compat import StreamTextType
+    from ruamel.yaml.nodes import ScalarNode
+    from ruamel.yaml.representer import RoundTripRepresenter
+    from ruamel.yaml.tokens import CommentToken
+
+    from ansiblelint.file_utils import Lintable
 
 _logger = logging.getLogger(__name__)
 
-YAMLLINT_CONFIG = """
-extends: default
-rules:
-  comments:
-    # https://github.com/prettier/prettier/issues/6780
-    min-spaces-from-content: 1
-  # https://github.com/adrienverge/yamllint/issues/384
-  comments-indentation: false
-  document-start: disable
-  # 160 chars was the default used by old E204 rule, but
-  # you can easily change it or disable in your .yamllint file.
-  line-length:
-    max: 160
-  # We are adding an extra space inside braces as that's how prettier does it
-  # and we are trying not to fight other linters.
-  braces:
-    min-spaces-inside: 0  # yamllint defaults to 0
-    max-spaces-inside: 1  # yamllint defaults to 0
-"""
+
+class CustomYamlLintConfig(YamlLintConfig):  # type: ignore[misc]
+    """Extension of YamlLintConfig."""
+
+    def __init__(
+        self,
+        content: str | None = None,
+        file: str | Path | None = None,
+    ) -> None:
+        """Initialize config."""
+        super().__init__(content, file)
+        self.incompatible = ""
 
 
 def deannotate(data: Any) -> Any:
@@ -85,88 +74,92 @@ def deannotate(data: Any) -> Any:
     return data
 
 
-@functools.lru_cache(maxsize=1)
-def load_yamllint_config() -> YamlLintConfig:
+def load_yamllint_config() -> CustomYamlLintConfig:
     """Load our default yamllint config and any customized override file."""
-    config = YamlLintConfig(content=YAMLLINT_CONFIG)
+    config = CustomYamlLintConfig(file=Path(__file__).parent / "data" / ".yamllint")
+    config.incompatible = ""
     # if we detect local yamllint config we use it but raise a warning
     # as this is likely to get out of sync with our internal config.
-    for file in [
+    for path in [
         ".yamllint",
         ".yamllint.yaml",
         ".yamllint.yml",
         os.getenv("YAMLLINT_CONFIG_FILE", ""),
-        os.getenv("XDG_CONFIG_HOME", os.path.expanduser("~/.config"))
-        + "/yamllint/config",
+        os.getenv("XDG_CONFIG_HOME", "~/.config") + "/yamllint/config",
     ]:
-        if os.path.isfile(file):
+        file = Path(path).expanduser()
+        if file.is_file():
             _logger.debug(
                 "Loading custom %s config file, this extends our "
                 "internal yamllint config.",
                 file,
             )
-            config_override = YamlLintConfig(file=file)
-            config_override.extend(config)
-            config = config_override
+            custom_config = CustomYamlLintConfig(file=str(file))
+            custom_config.extend(config)
+            config = custom_config
             break
+
+    # Look for settings incompatible with our reformatting
+    checks: list[tuple[str, str | int | bool]] = [
+        (
+            "comments.min-spaces-from-content",
+            1,
+        ),
+        (
+            "comments-indentation",
+            False,
+        ),
+        (
+            "braces.min-spaces-inside",
+            0,
+        ),
+        (
+            "braces.max-spaces-inside",
+            1,
+        ),
+        (
+            "octal-values.forbid-implicit-octal",
+            True,
+        ),
+        (
+            "octal-values.forbid-explicit-octal",
+            True,
+        ),
+        # (
+        #     "key-duplicates.forbid-duplicated-merge-keys", # v1.34.0+
+        #     True,
+        # ),
+        # (
+        #   "quoted-strings.quote-type", "double",
+        # ),
+        # (
+        #   "quoted-strings.required", "only-when-needed",
+        # ),
+    ]
+    errors = []
+    for setting, expected_value in checks:
+        v = config.rules
+        for key in setting.split("."):
+            if not isinstance(v, dict):  # pragma: no cover
+                break
+            if key not in v:  # pragma: no cover
+                break
+            v = v[key]
+        if v != expected_value:
+            msg = f"{setting} must be {str(expected_value).lower()}"
+            errors.append(msg)
+    if errors:
+        nl = "\n"
+        msg = f"Found incompatible custom yamllint configuration ({file}), please either remove the file or edit it to comply with:{nl}  - {(nl + '  - ').join(errors)}.{nl}{nl}Read https://ansible.readthedocs.io/projects/lint/rules/yaml/ for more details regarding why we have these requirements. Fix mode will not be available."
+        config.incompatible = msg
+
     _logger.debug("Effective yamllint rules used: %s", config.rules)
     return config
 
 
-def iter_tasks_in_file(
-    lintable: Lintable,
-) -> Iterator[tuple[dict[str, Any], dict[str, Any], list[str], MatchError | None]]:
-    """Iterate over tasks in file.
-
-    This yields a 4-tuple of raw_task, normalized_task, skip_tags, and error.
-
-    raw_task:
-        When looping through the tasks in the file, each "raw_task" is minimally
-        processed to include these special keys: __line__, __file__, skipped_rules.
-    normalized_task:
-        When each raw_task is "normalized", action shorthand (strings) get parsed
-        by ansible into python objects and the action key gets normalized. If the task
-        should be skipped (skipped is True) or normalizing it fails (error is not None)
-        then this is just the raw_task instead of a normalized copy.
-    skip_tags:
-        List of tags found to be skipped, from tags block or noqa comments
-    error:
-        This is normally None. It will be a MatchError when the raw_task cannot be
-        normalized due to an AnsibleParserError.
-
-    :param lintable: The playbook or tasks/handlers yaml file to get tasks from
-
-    Yields raw_task, normalized_task, skipped, error
-    """
-    data = lintable.data
-    if not data:
-        return
-
-    raw_tasks = get_action_tasks(data, lintable)
-
-    for raw_task in raw_tasks:
-        err: MatchError | None = None
-
-        skip_tags: list[str] = raw_task.get(SKIPPED_RULES_KEY, [])
-
-        try:
-            normalized_task = normalize_task(raw_task, str(lintable.path))
-        except MatchError as err:
-            # normalize_task converts AnsibleParserError to MatchError
-            yield raw_task, raw_task, skip_tags, err
-            return
-
-        if "skip_ansible_lint" in raw_task.get("tags", []):
-            skip_tags.append("skip_ansible_lint")
-        if skip_tags:
-            yield raw_task, normalized_task, skip_tags, err
-            continue
-
-        yield raw_task, normalized_task, skip_tags, err
-
-
 def nested_items_path(
     data_collection: dict[Any, Any] | list[Any],
+    ignored_keys: Sequence[str] = (),
 ) -> Iterator[tuple[Any, Any, list[str | int]]]:
     """Iterate a nested data structure, yielding key/index, value, and parent_path.
 
@@ -227,12 +220,22 @@ def nested_items_path(
     # valid data, we better ignore NoneType
     if data_collection is None:
         return
-    yield from _nested_items_path(data_collection=data_collection, parent_path=[])
+    data: dict[Any, Any] | list[Any]
+    if isinstance(data_collection, Task):
+        data = data_collection.normalized_task
+    else:
+        data = data_collection
+    yield from _nested_items_path(
+        data_collection=data,
+        parent_path=[],
+        ignored_keys=ignored_keys,
+    )
 
 
 def _nested_items_path(
     data_collection: dict[Any, Any] | list[Any],
     parent_path: list[str | int],
+    ignored_keys: Sequence[str] = (),
 ) -> Iterator[tuple[Any, Any, list[str | int]]]:
     """Iterate through data_collection (internal implementation of nested_items_path).
 
@@ -242,43 +245,45 @@ def _nested_items_path(
     """
     # we have to cast each convert_to_tuples assignment or mypy complains
     # that both assignments (for dict and list) do not have the same type
-    convert_to_tuples_type = Callable[[], Iterator[Tuple[Union[str, int], Any]]]
+    convert_to_tuples_type = Callable[[], Iterator[tuple[str | int, Any]]]
     if isinstance(data_collection, dict):
         convert_data_collection_to_tuples = cast(
-            convert_to_tuples_type, functools.partial(data_collection.items)
+            convert_to_tuples_type,
+            functools.partial(data_collection.items),
         )
     elif isinstance(data_collection, list):
         convert_data_collection_to_tuples = cast(
-            convert_to_tuples_type, functools.partial(enumerate, data_collection)
+            convert_to_tuples_type,
+            functools.partial(enumerate, data_collection),
         )
     else:
-        raise TypeError(
-            f"Expected a dict or a list but got {data_collection!r} "
-            f"of type '{type(data_collection)}'"
-        )
+        msg = f"Expected a dict or a list but got {data_collection!r} of type '{type(data_collection)}'"
+        raise TypeError(msg)
     for key, value in convert_data_collection_to_tuples():
-        if key in (SKIPPED_RULES_KEY, "__file__", "__line__"):
+        if key in (*ANNOTATION_KEYS, *ignored_keys):
             continue
         yield key, value, parent_path
-        if isinstance(value, (dict, list)):
+        if isinstance(value, dict | list):
             yield from _nested_items_path(
-                data_collection=value, parent_path=parent_path + [key]
+                data_collection=value,
+                parent_path=[*parent_path, key],
             )
 
 
 def get_path_to_play(
     lintable: Lintable,
-    line_number: int,  # 1-based
+    lineno: int,  # 1-based
     ruamel_data: CommentedMap | CommentedSeq,
 ) -> list[str | int]:
     """Get the path to the play in the given file at the given line number."""
-    if line_number < 1:
-        raise ValueError(f"expected line_number >= 1, got {line_number}")
+    if lineno < 1:
+        msg = f"expected lineno >= 1, got {lineno}"
+        raise ValueError(msg)
     if lintable.kind != "playbook" or not isinstance(ruamel_data, CommentedSeq):
         return []
-    lc: LineCol  # lc uses 0-based counts # pylint: disable=invalid-name
-    # line_number is 1-based. Convert to 0-based.
-    line_index = line_number - 1
+    lc: LineCol  # lc uses 0-based counts
+    # lineno is 1-based. Convert to 0-based.
+    line_index = lineno - 1
 
     prev_play_line_index = ruamel_data.lc.line
     last_play_index = len(ruamel_data)
@@ -289,8 +294,10 @@ def get_path_to_play(
         else:
             next_play_line_index = None
 
-        lc = play.lc  # pylint: disable=invalid-name
-        assert isinstance(lc.line, int)
+        lc = play.lc
+        if not isinstance(lc.line, int):
+            msg = f"expected lc.line to be an int, got {lc.line!r}"
+            raise TypeError(msg)
         if lc.line == line_index:
             return [play_index]
         if play_index > 0 and prev_play_line_index < line_index < lc.line:
@@ -310,25 +317,27 @@ def get_path_to_play(
 
 def get_path_to_task(
     lintable: Lintable,
-    line_number: int,  # 1-based
+    lineno: int,  # 1-based
     ruamel_data: CommentedMap | CommentedSeq,
 ) -> list[str | int]:
     """Get the path to the task in the given file at the given line number."""
-    if line_number < 1:
-        raise ValueError(f"expected line_number >= 1, got {line_number}")
-    if lintable.kind in ("tasks", "handlers"):
-        assert isinstance(ruamel_data, CommentedSeq)
-        return _get_path_to_task_in_tasks_block(line_number, ruamel_data)
-    if lintable.kind == "playbook":
-        assert isinstance(ruamel_data, CommentedSeq)
-        return _get_path_to_task_in_playbook(line_number, ruamel_data)
-    # if lintable.kind in ["yaml", "requirements", "vars", "meta", "reno", "test-meta"]:
+    if lineno < 1:
+        msg = f"expected lineno >= 1, got {lineno}"
+        raise ValueError(msg)
+    if lintable.kind in ("tasks", "handlers", "playbook"):
+        if not isinstance(ruamel_data, CommentedSeq):
+            msg = f"expected ruamel_data to be a CommentedSeq, got {ruamel_data!r}"
+            raise ValueError(msg)
+        if lintable.kind in ("tasks", "handlers"):
+            return _get_path_to_task_in_tasks_block(lineno, ruamel_data)
+        if lintable.kind == "playbook":
+            return _get_path_to_task_in_playbook(lineno, ruamel_data)
 
     return []
 
 
 def _get_path_to_task_in_playbook(
-    line_number: int,  # 1-based
+    lineno: int,  # 1-based
     ruamel_data: CommentedSeq,
 ) -> list[str | int]:
     """Get the path to the task in the given playbook data at the given line number."""
@@ -340,6 +349,10 @@ def _get_path_to_task_in_playbook(
         else:
             next_play_line_index = None
 
+        # We clearly haven't found the right spot yet if a following play starts on an earlier line.
+        if next_play_line_index and lineno > next_play_line_index:
+            continue
+
         play_keys = list(play.keys())
         for tasks_keyword in PLAYBOOK_TASK_KEYWORDS:
             if not play.get(tasks_keyword):
@@ -351,19 +364,20 @@ def _get_path_to_task_in_playbook(
                 next_block_line_index = None
             else:
                 next_block_line_index = play.lc.data[next_keyword][0]
-            # last_line_number_in_block is 1-based; next_*_line_index is 0-based
+            # last_lineno_in_block is 1-based; next_*_line_index is 0-based
             # next_*_line_index - 1 to get line before next_*_line_index.
             # Then + 1 to make it a 1-based number.
-            # So, last_line_number_in_block = next_*_line_index - 1 + 1
             if next_block_line_index is not None:
-                last_line_number_in_block = next_block_line_index
+                last_lineno_in_block = next_block_line_index
             elif next_play_line_index is not None:
-                last_line_number_in_block = next_play_line_index
+                last_lineno_in_block = next_play_line_index
             else:
-                last_line_number_in_block = None
+                last_lineno_in_block = None
 
             task_path = _get_path_to_task_in_tasks_block(
-                line_number, play[tasks_keyword], last_line_number_in_block
+                lineno,
+                play[tasks_keyword],
+                last_lineno_in_block,
             )
             if task_path:
                 # mypy gets confused without this typehint
@@ -372,20 +386,20 @@ def _get_path_to_task_in_playbook(
                     tasks_keyword,
                 ]
                 return tasks_keyword_path + list(task_path)
-    # line_number is before first play or no tasks keywords in any of the plays
+    # lineno is before first play or no tasks keywords in any of the plays
     return []
 
 
 def _get_path_to_task_in_tasks_block(
-    line_number: int,  # 1-based
+    lineno: int,  # 1-based
     tasks_block: CommentedSeq,
-    last_line_number: int | None = None,  # 1-based
+    last_lineno: int | None = None,  # 1-based
 ) -> list[str | int]:
     """Get the path to the task in the given tasks block at the given line number."""
     task: CommentedMap | None
-    # line_number and last_line_number are 1-based. Convert to 0-based.
-    line_index = line_number - 1
-    last_line_index = None if last_line_number is None else last_line_number - 1
+    # lineno and last_lineno are 1-based. Convert to 0-based.
+    line_index = lineno - 1
+    last_line_index = None if last_lineno is None else last_lineno - 1
 
     # lc (LineCol) uses 0-based counts
     prev_task_line_index = tasks_block.lc.line
@@ -408,14 +422,19 @@ def _get_path_to_task_in_tasks_block(
         nested_task_keys = set(task.keys()).intersection(set(NESTED_TASK_KEYS))
         if nested_task_keys:
             subtask_path = _get_path_to_task_in_nested_tasks_block(
-                line_number, task, nested_task_keys, next_task_line_index
+                lineno,
+                task,
+                nested_task_keys,
+                next_task_line_index,
             )
             if subtask_path:
                 # mypy gets confused without this typehint
                 task_path: list[str | int] = [task_index]
                 return task_path + list(subtask_path)
 
-        assert isinstance(task.lc.line, int)
+        if not isinstance(task.lc.line, int):
+            msg = f"expected task.lc.line to be an int, got {task.lc.line!r}"
+            raise TypeError(msg)
         if task.lc.line == line_index:
             return [task_index]
         if task_index > 0 and prev_task_line_index < line_index < task.lc.line:
@@ -437,7 +456,7 @@ def _get_path_to_task_in_tasks_block(
 
 
 def _get_path_to_task_in_nested_tasks_block(
-    line_number: int,  # 1-based
+    lineno: int,  # 1-based
     task: CommentedMap,
     nested_task_keys: set[str],
     next_task_line_index: int | None = None,  # 0-based
@@ -452,25 +471,26 @@ def _get_path_to_task_in_nested_tasks_block(
             continue
         next_task_key = task_keys_by_index.get(task_index + 1, None)
         if next_task_key is not None:
+            if task.lc.data[next_task_key][2] < lineno:
+                continue
             next_task_key_line_index = task.lc.data[next_task_key][0]
         else:
             next_task_key_line_index = None
-        # last_line_number_in_block is 1-based; next_*_line_index is 0-based
+        # last_lineno_in_block is 1-based; next_*_line_index is 0-based
         # next_*_line_index - 1 to get line before next_*_line_index.
         # Then + 1 to make it a 1-based number.
-        # So, last_line_number_in_block = next_*_line_index - 1 + 1
-        last_line_number_in_block = (
+        last_lineno_in_block = (
             next_task_key_line_index
             if next_task_key_line_index is not None
             else next_task_line_index
         )
         subtask_path = _get_path_to_task_in_tasks_block(
-            line_number,
+            lineno,
             nested_task_block,
-            last_line_number_in_block,  # 1-based
+            last_lineno_in_block,  # 1-based
         )
         if subtask_path:
-            return [task_key] + list(subtask_path)
+            return [task_key, *list(subtask_path)]
     # line is not part of this nested tasks block
     return []
 
@@ -496,8 +516,12 @@ class OctalIntYAML11(ScalarInt):
         v = format(data, "o")
         anchor = data.yaml_anchor(any=True)
         # noinspection PyProtectedMember
-        # pylint: disable=protected-access
-        return representer.insert_underscore("0", v, data._underscore, anchor=anchor)
+        return representer.insert_underscore(
+            "0",
+            v,
+            data._underscore,  # noqa: SLF001
+            anchor=anchor,
+        )
 
 
 class CustomConstructor(RoundTripConstructor):
@@ -528,16 +552,22 @@ class CustomConstructor(RoundTripConstructor):
             value_s = value_su.replace("_", "")
             if value_s[0] in "+-":
                 value_s = value_s[1:]
-            if value_s[0] == "0":
+            if value_s[0:2] == "0x":
+                ret = HexInt(ret, width=len(value_s) - 2)
+            elif value_s[0] == "0":
                 # got an octal in YAML 1.1
                 ret = OctalIntYAML11(
-                    ret, width=None, underscore=underscore, anchor=node.anchor
+                    ret,
+                    width=None,
+                    underscore=underscore,
+                    anchor=node.anchor,
                 )
         return ret
 
 
 CustomConstructor.add_constructor(
-    "tag:yaml.org,2002:int", CustomConstructor.construct_yaml_int
+    "tag:yaml.org,2002:int",
+    CustomConstructor.construct_yaml_int,
 )
 
 
@@ -575,7 +605,8 @@ class FormattedEmitter(Emitter):
     def expect_document_root(self) -> None:
         """Expect doc root (extend to record if the root doc is a sequence)."""
         self._root_is_sequence = isinstance(
-            self.event, ruamel.yaml.events.SequenceStartEvent
+            self.event,
+            ruamel.yaml.events.SequenceStartEvent,
         )
         return super().expect_document_root()
 
@@ -606,6 +637,34 @@ class FormattedEmitter(Emitter):
     def choose_scalar_style(self) -> Any:
         """Select how to quote scalars if needed."""
         style = super().choose_scalar_style()
+        if (
+            style == ""
+            and self.event.value.startswith("0")
+            and len(self.event.value) > 1
+        ):
+            # We have an as-yet unquoted token that starts with "0" (but is not itself the digit 0).
+            # It could be:
+            # - hexadecimal like "0xF1" or ; comes tagged as int. Should continue unquoted to continue as an int.
+            # - octal like "0666" or "0o755"; comes tagged as str. **Should** be quoted to be cross-YAML compatible.
+            # - string like "0.0.0.0" and "00-header". Should not be quoted, unless it has a quote in it.
+            if (
+                self.event.value.startswith("0x")
+                and self.event.tag == "tag:yaml.org,2002:int"
+                and self.event.implicit[0]
+            ):
+                # hexadecimal
+                self.event.tag = "tag:yaml.org,2002:str"
+                return ""
+            try:
+                int(self.event.value, 8)
+            except ValueError:
+                pass
+                # fallthrough to string
+            else:
+                # octal
+                self.event.tag = "tag:yaml.org,2002:str"
+                self.event.implicit = (True, True, True)
+                return '"'
         if style != "'":
             # block scalar, double quoted, etc.
             return style
@@ -613,12 +672,23 @@ class FormattedEmitter(Emitter):
             return "'"
         return self.preferred_quote
 
+    def increase_indent(
+        self,
+        flow: bool = False,  # noqa: FBT002
+        sequence: bool | None = None,
+        indentless: bool = False,  # noqa: FBT002
+    ) -> None:
+        super().increase_indent(flow, sequence, indentless)
+        # If our previous node was a sequence and we are still trying to indent, don't
+        if self.indents.last_seq():
+            self.indent = self.column + 1
+
     def write_indicator(
         self,
         indicator: str,  # ruamel.yaml typehint is wrong. This is a string.
         need_whitespace: bool,
-        whitespace: bool = False,
-        indention: bool = False,  # (sic) ruamel.yaml has this typo in their API
+        whitespace: bool = False,  # noqa: FBT002
+        indention: bool = False,  # (sic) ruamel.yaml has this typo in their API # noqa: FBT002
     ) -> None:
         """Make sure that flow maps get whitespace by the curly braces."""
         # We try to go with one whitespace by the curly braces and adjust accordingly
@@ -635,6 +705,9 @@ class FormattedEmitter(Emitter):
             and not self._in_empty_flow_map
         ):
             indicator = (" " * spaces_inside) + "}"
+        # Indicator sometimes comes with embedded spaces we need to squish
+        if indicator == "  -" and self.indents.last_seq():
+            indicator = "-"
         super().write_indicator(indicator, need_whitespace, whitespace, indention)
         # if it is the start of a flow mapping, and it's not time
         # to wrap the lines, insert a space.
@@ -655,7 +728,7 @@ class FormattedEmitter(Emitter):
         """Modify strings to protect "#" from full-line-comment post-processing."""
         try:
             if "#" in string:
-                # ＃ is \uFF03 (fullwidth number sign)
+                # # is \uFF03 (fullwidth number sign)
                 # ﹟ is \uFE5F (small number sign)
                 string = string.replace("#", "\uFF03#\uFE5F")
                 # this is safe even if this sequence is present
@@ -670,7 +743,7 @@ class FormattedEmitter(Emitter):
         """Remove string protection of "#" after full-line-comment post-processing."""
         try:
             if "\uFF03#\uFE5F" in string:
-                # ＃ is \uFF03 (fullwidth number sign)
+                # # is \uFF03 (fullwidth number sign)
                 # ﹟ is \uFE5F (small number sign)
                 string = string.replace("\uFF03#\uFE5F", "#")
         except (ValueError, TypeError):
@@ -690,7 +763,11 @@ class FormattedEmitter(Emitter):
         return analysis
 
     # comment is a CommentToken, not Any (Any is ruamel.yaml's lazy type hint).
-    def write_comment(self, comment: CommentToken, pre: bool = False) -> None:
+    def write_comment(
+        self,
+        comment: CommentToken,
+        pre: bool = False,  # noqa: FBT002
+    ) -> None:
         """Clean up extra new lines and spaces in comments.
 
         ruamel.yaml treats new or empty lines as comments.
@@ -702,16 +779,21 @@ class FormattedEmitter(Emitter):
             and not value.strip()
             and not isinstance(
                 self.event,
-                (
-                    ruamel.yaml.events.CollectionEndEvent,
-                    ruamel.yaml.events.DocumentEndEvent,
-                    ruamel.yaml.events.StreamEndEvent,
-                ),
+                ruamel.yaml.events.CollectionEndEvent
+                | ruamel.yaml.events.DocumentEndEvent
+                | ruamel.yaml.events.StreamEndEvent
+                | ruamel.yaml.events.MappingStartEvent,
             )
         ):
             # drop pure whitespace pre comments
             # does not apply to End events since they consume one of the newlines.
             value = ""
+        elif (
+            pre
+            and not value.strip()
+            and isinstance(self.event, ruamel.yaml.events.MappingStartEvent)
+        ):
+            value = self._re_repeat_blank_lines.sub("", value)
         elif pre:
             # preserve content in pre comment with at least one newline,
             # but no extra blank lines.
@@ -738,14 +820,25 @@ class FormattedEmitter(Emitter):
 class FormattedYAML(YAML):
     """A YAML loader/dumper that handles ansible content better by default."""
 
-    def __init__(
+    default_config = {
+        "explicit_start": True,
+        "explicit_end": False,
+        "width": 160,
+        "indent_sequences": True,
+        "preferred_quote": '"',
+        "min_spaces_inside": 0,
+        "max_spaces_inside": 1,
+    }
+
+    def __init__(  # pylint: disable=too-many-arguments
         self,
         *,
         typ: str | None = None,
         pure: bool = False,
         output: Any = None,
-        # input: Any = None,
         plug_ins: list[str] | None = None,
+        version: tuple[int, int] | None = None,
+        config: dict[str, bool | int | str] | None = None,
     ):
         """Return a configured ``ruamel.yaml.YAML`` instance.
 
@@ -805,15 +898,18 @@ class FormattedYAML(YAML):
               tasks:
                 - name: Task
         """
-        # Default to reading/dumping YAML 1.1 (ruamel.yaml defaults to 1.2)
-        self._yaml_version_default: tuple[int, int] = (1, 1)
-        self._yaml_version: str | tuple[int, int] = self._yaml_version_default
-
+        if version:
+            if isinstance(version, str):
+                x, y = version.split(".", maxsplit=1)
+                version = (int(x), int(y))
+            self._yaml_version_default: tuple[int, int] = version
+            self._yaml_version: tuple[int, int] = self._yaml_version_default
         super().__init__(typ=typ, pure=pure, output=output, plug_ins=plug_ins)
 
         # NB: We ignore some mypy issues because ruamel.yaml typehints are not great.
 
-        config = self._defaults_from_yamllint_config()
+        if not config:
+            config = self._defaults_from_yamllint_config()
 
         # these settings are derived from yamllint config
         self.explicit_start: bool = config["explicit_start"]  # type: ignore[assignment]
@@ -830,9 +926,9 @@ class FormattedYAML(YAML):
         self.compact_seq_map = True  # type: ignore[assignment] # key after dash
 
         # Do not use yaml.indent() as it obscures the purpose of these vars:
-        self.map_indent = 2  # type: ignore[assignment]
-        self.sequence_indent = 4 if indent_sequences else 2  # type: ignore[assignment]
-        self.sequence_dash_offset = self.sequence_indent - 2  # type: ignore[operator]
+        self.map_indent = 2
+        self.sequence_indent = 4 if indent_sequences else 2
+        self.sequence_dash_offset = self.sequence_indent - 2
 
         # If someone doesn't want our FormattedEmitter, they can change it.
         self.Emitter = FormattedEmitter
@@ -858,27 +954,16 @@ class FormattedYAML(YAML):
         # This will only preserve quotes for strings read from the file.
         # anything modified by the transform will use no quotes, preferred_quote,
         # or the quote that results in the least amount of escaping.
-        # self.preserve_quotes = True
 
         # If needed, we can use this to change null representation to be explicit
         # (see https://stackoverflow.com/a/44314840/1134951)
         # self.Representer.add_representer(
-        #     type(None),
-        #     lambda self, data: self.represent_scalar("tag:yaml.org,2002:null", "null"),
-        # )
 
     @staticmethod
     def _defaults_from_yamllint_config() -> dict[str, bool | int | str]:
         """Extract FormattedYAML-relevant settings from yamllint config if possible."""
-        config = {
-            "explicit_start": True,
-            "explicit_end": False,
-            "width": 160,
-            "indent_sequences": True,
-            "preferred_quote": '"',
-            "min_spaces_inside": 0,
-            "max_spaces_inside": 1,
-        }
+        config = FormattedYAML.default_config
+
         for rule, rule_config in load_yamllint_config().rules.items():
             if not rule_config:
                 # rule disabled
@@ -911,10 +996,10 @@ class FormattedYAML(YAML):
                 elif quote_type == "double":
                     config["preferred_quote"] = '"'
 
-        return cast(Dict[str, Union[bool, int, str]], config)
+        return cast(dict[str, bool | int | str], config)
 
-    @property  # type: ignore[override]
-    def version(self) -> str | tuple[int, int]:
+    @property
+    def version(self) -> tuple[int, int] | None:
         """Return the YAML version used to parse or dump.
 
         Ansible uses PyYAML which only supports YAML 1.1. ruamel.yaml defaults to 1.2.
@@ -922,26 +1007,51 @@ class FormattedYAML(YAML):
         We can relax the version requirement once ansible uses a version of PyYAML
         that includes this PR: https://github.com/yaml/pyyaml/pull/555
         """
-        return self._yaml_version
+        if hasattr(self, "_yaml_version"):
+            return self._yaml_version
+        return None
 
     @version.setter
-    def version(self, value: str | tuple[int, int] | None) -> None:
+    def version(self, value: tuple[int, int] | None) -> None:
         """Ensure that yaml version uses our default value.
 
         The yaml Reader updates this value based on the ``%YAML`` directive in files.
         So, if a file does not include the directive, it sets this to None.
         But, None effectively resets the parsing version to YAML 1.2 (ruamel's default).
         """
-        self._yaml_version = value if value is not None else self._yaml_version_default
+        if value is not None:
+            self._yaml_version = value
+        elif hasattr(self, "_yaml_version_default"):
+            self._yaml_version = self._yaml_version_default
+        # We do nothing if the object did not have a previous default version defined
 
-    def loads(self, stream: str) -> Any:
+    def load(self, stream: Path | StreamTextType) -> Any:
         """Load YAML content from a string while avoiding known ruamel.yaml issues."""
         if not isinstance(stream, str):
-            raise NotImplementedError(f"expected a str but got {type(stream)}")
+            msg = f"expected a str but got {type(stream)}"
+            raise NotImplementedError(msg)
+        # As ruamel drops comments for any document that is not a mapping or sequence,
+        # we need to avoid using it to reformat those documents.
+        # https://sourceforge.net/p/ruamel-yaml/tickets/460/
+
         text, preamble_comment = self._pre_process_yaml(stream)
-        data = self.load(stream=text)
-        if preamble_comment is not None:
-            setattr(data, "preamble_comment", preamble_comment)
+        try:
+            data = super().load(stream=text)
+        except ComposerError:
+            data = self.load_all(stream=text)
+        except ParserError:
+            data = None
+            _logger.error(  # noqa: TRY400
+                "Invalid yaml, verify the file contents and try again.",
+            )
+        if preamble_comment is not None and isinstance(
+            data,
+            CommentedMap | CommentedSeq,
+        ):
+            data.preamble_comment = preamble_comment  # type: ignore[union-attr]
+        # Because data can validly also be None for empty documents, we cannot
+        # really annotate the return type here, so we need to remember to
+        # never save None or scalar data types when reformatting.
         return data
 
     def dumps(self, data: Any) -> str:
@@ -953,15 +1063,20 @@ class FormattedYAML(YAML):
                 stream.write(preamble_comment)
             self.dump(data, stream)
             text = stream.getvalue()
-        return self._post_process_yaml(text)
+        strip_version_directive = hasattr(self, "_yaml_version_default")
+        return self._post_process_yaml(
+            text,
+            strip_version_directive=strip_version_directive,
+            strip_explicit_start=not self.explicit_start,
+        )
 
     def _prevent_wrapping_flow_style(self, data: Any) -> None:
-        if not isinstance(data, (CommentedMap, CommentedSeq)):
+        if not isinstance(data, CommentedMap | CommentedSeq):
             return
         for key, value, parent_path in nested_items_path(data):
-            if not isinstance(value, (CommentedMap, CommentedSeq)):
+            if not isinstance(value, CommentedMap | CommentedSeq):
                 continue
-            fa: Format = value.fa  # pylint: disable=invalid-name
+            fa: Format = value.fa
             if fa.flow_style():
                 predicted_indent = self._predict_indent_length(parent_path, key)
                 predicted_width = len(str(value))
@@ -1017,14 +1132,11 @@ class FormattedYAML(YAML):
         #     DocumentStartToken.comment -> DocumentStartEvent.comment
         #   Then, in the composer:
         #     once in composer.current_event
-        #       composer.compose_document()
         #         discards DocumentStartEvent
         #           move DocumentStartEvent to composer.last_event
-        #           node = composer.compose_node(None, None)
         #             all document nodes get composed (events get used)
         #         discard DocumentEndEvent
         #           move DocumentEndEvent to composer.last_event
-        #         return node
         # So, there's no convenient way to extend the composer
         # to somehow capture the comments and pass them on.
 
@@ -1044,7 +1156,12 @@ class FormattedYAML(YAML):
         return text, "".join(preamble_comments) or None
 
     @staticmethod
-    def _post_process_yaml(text: str) -> str:
+    def _post_process_yaml(
+        text: str,
+        *,
+        strip_version_directive: bool = False,
+        strip_explicit_start: bool = False,
+    ) -> str:
         """Handle known issues with ruamel.yaml dumping.
 
         Make sure there's only one newline at the end of the file.
@@ -1056,6 +1173,14 @@ class FormattedYAML(YAML):
 
         Make sure null list items don't end in a space.
         """
+        # remove YAML directive
+        if strip_version_directive and text.startswith("%YAML"):
+            text = text.split("\n", 1)[1]
+
+        # remove explicit document start
+        if strip_explicit_start and text.startswith("---"):
+            text = text.split("\n", 1)[1]
+
         text = text.rstrip("\n") + "\n"
 
         lines = text.splitlines(keepends=True)
@@ -1100,12 +1225,11 @@ class FormattedYAML(YAML):
 
 def clean_json(
     obj: Any,
-    func: Callable[[str], Any] = lambda key: key.startswith("__")
-    if isinstance(key, str)
-    else False,
+    func: Callable[[str], Any] = lambda key: (
+        key.startswith("__") if isinstance(key, str) else False
+    ),
 ) -> Any:
-    """
-    Remove all keys matching the condition from a nested JSON-like object.
+    """Remove all keys matching the condition from a nested JSON-like object.
 
     :param obj: a JSON like object to clean, also returned for chaining.
     :param func: a callable that takes a key in argument and return True for each key to delete
